import json
import time

import pytest

IN_DEV = True


class TestSchemaMutations:
    create_pipeline_mutation = """
        mutation CreatePipeline($pipeline: PipelineInput!, $uniquePaths: [String!]) {
          createPipeline(pipeline: $pipeline, uniquePaths: $uniquePaths) {
              id
              name
              describe
              nodes {
                name
                inputs
                outputs
                tags
              }
              parameters {
                name
                value
                type
              }
              status {
                finishedAt
                state
                runner
                session
                startedAt
                taskArgs
                taskEinfo
                taskException
                taskId
                taskKwargs
                taskName
                taskRequest
                taskResult
                taskTraceback
                filteredNodes
              }
              tags {
                key
                value
              }
              parent
              createdAt
              dataCatalog {
                name
                config
              }
          }
        }
        """

    update_pipeline_mutation = """
        mutation UpdatePipeline($pipeline: PipelineInput!, $id: String!, $uniquePaths: [String!]) {
          updatePipeline(pipeline: $pipeline, id: $id, uniquePaths: $uniquePaths) {
            id
            name
            describe
            nodes {
              name
              inputs
              outputs
              tags
            }
            parameters {
              name
              value
              type
            }
            status {
              finishedAt
              state
              runner
              session
              startedAt
              taskArgs
              taskEinfo
              taskException
              taskId
              taskKwargs
              taskName
              taskRequest
              taskResult
              taskTraceback
              filteredNodes
            }
            tags {
              key
              value
            }
            parent
            createdAt
            dataCatalog {
                name
                config
            }
          }
        }
        """

    delete_pipeline_mutation = """
        mutation TestMutation($id: String!) {
          deletePipeline(id: $id) {
           id
            name
            describe
            nodes {
              name
              inputs
              outputs
              tags
            }
            parameters {
              name
              value
              type
            }
            status {
              finishedAt
              state
              runner
              session
              startedAt
              taskArgs
              taskEinfo
              taskException
              taskId
              taskKwargs
              taskName
              taskRequest
              taskResult
              taskTraceback
              filteredNodes
            }
            tags {
              key
              value
            }
            parent
            createdAt
            dataCatalog {
                name
                config
            }
          }
        }
        """

    create_datasets_mutation = """
        mutation CreateDatasets($id: String!, $datasets: [DataSetInput!]!, $expiresInSec: Int!) {
          createDatasets(id: $id, datasets: $datasets, expiresInSec: $expiresInSec){
            __typename
            ... on SignedUrl {
              url
              file
              fields {
                name
                value
              }
            }
            ... on SignedUrls {
              urls {
                url
                file
                fields {
                  name
                  value
                }
              }
            }
          }
        }
        """

    @pytest.mark.asyncio
    async def test_create_pipeline_00(self,
                                      mock_app,
                                      mock_celery_session_app,
                                      celery_session_worker,
                                      mock_info_context,
                                      mock_text_in,
                                      mock_text_out):

        resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                 {"name": "text_out", "config": json.dumps(
                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                             },
                                                 "uniquePaths": ["text_in", "text_out"]})

        assert resp.errors is None
        # check filepaths are unique
        datasets = {d["name"]: json.loads(d["config"])
                    for d in resp.data["createPipeline"]["dataCatalog"]}
        assert datasets["text_in"]["filepath"] != str(mock_text_in)
        assert datasets["text_out"]["filepath"] != str(mock_text_out)
        # making sure the uniquePaths parameter is working
        assert str(resp.data["createPipeline"]["id"]
                   ) == datasets["text_in"]["filepath"].rsplit("/", 2)[1]

    @pytest.mark.asyncio
    async def test_create_pipeline_01(self,
                                      mock_app,
                                      mock_celery_session_app,
                                      celery_session_worker,
                                      mock_info_context,
                                      mock_text_in_tsv,
                                      mock_text_out_tsv):

        resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                             variable_values={"pipeline": {
                                                 "name": "example00",
                                                 "dataCatalog": [
                                                     {"name": "text_in", "config": json.dumps({"type": "pandas.CSVDataset", "filepath": str(mock_text_in_tsv), "loadArgs": [
                                                                                              {"name": "sep", "value": "\t"}], "saveArgs": [{"name": "sep", "value": "\t"}]})},
                                                     {"name": "text_out", "config": json.dumps({"type": "pandas.CSVDataset", "filepath": str(
                                                         mock_text_out_tsv), "loadArgs": [{"name": "sep", "value": "\t"}], "saveArgs": [{"name": "sep", "value": "\t"}]})}
                                                 ],
                                                 "parameters": [{"name": "example", "value": "hello"}],
                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                             },
                                                 "uniquePaths": None})

        assert resp.errors is None

    @pytest.mark.asyncio
    async def test_create_pipeline_staged(self,
                                          mock_app,
                                          mock_celery_session_app,
                                          celery_session_worker,
                                          mock_info_context,
                                          mock_text_in,
                                          mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "STAGED",
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             },
                                                                 "uniquePaths": None})

        pipeline_state = create_pipeline_resp.data["createPipeline"]["status"][-1]["state"]

        assert create_pipeline_resp.errors is None
        assert pipeline_state == "STAGED"

    @pytest.mark.asyncio
    async def test_create_pipeline_ready(self,
                                         mock_app,
                                         mock_celery_session_app,
                                         celery_session_worker,
                                         mock_info_context,
                                         mock_text_in,
                                         mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             },
                                                                 "uniquePaths": None})

        pipeline_state = create_pipeline_resp.data["createPipeline"]["status"][-1]["state"]

        assert create_pipeline_resp.errors is None
        assert pipeline_state != "STAGED"

    @pytest.mark.asyncio
    async def test_create_pipeline_ready_invalid_name(self,
                                                      mock_app,
                                                      mock_celery_session_app,
                                                      celery_session_worker,
                                                      mock_info_context,
                                                      mock_text_in,
                                                      mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example02",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             },
                                                                 "uniquePaths": None})

        assert create_pipeline_resp.errors is not None

    @pytest.mark.asyncio
    async def test_update_pipeline_staged_to_ready(self,
                                                   mock_app,
                                                   mock_celery_session_app,
                                                   celery_session_worker,
                                                   mock_info_context,
                                                   mock_text_in,
                                                   mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "state": "STAGED",
                                                             },
                                                                 "uniquePaths": None
                                                             })

        pipeline_id = create_pipeline_resp.data["createPipeline"]["id"]

        update_pipeline_resp = await mock_app.schema.execute(self.update_pipeline_mutation,
                                                             variable_values={"id": pipeline_id,
                                                                              "pipeline": {
                                                                                  "name": "example00",
                                                                                  "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                                  {"name": "text_out", "config": json.dumps(
                                                                                                      {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                                  ],
                                                                                  "parameters": [{"name": "example", "value": "hello"},
                                                                                                 {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                                  "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}],
                                                                                  "state": "READY",
                                                                              },
                                                                              "uniquePaths": None
                                                                              })
        assert update_pipeline_resp.errors is None
        assert update_pipeline_resp.data["updatePipeline"]["status"][-1]["state"] != "STAGED"
        assert update_pipeline_resp.data["updatePipeline"]["dataCatalog"][0] == {
            "name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})}
        assert update_pipeline_resp.data["updatePipeline"]["dataCatalog"][1] == {
            "name": "text_out", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_out)})}
        assert update_pipeline_resp.data["updatePipeline"]["parameters"][0] == {
            "name": "example", "value": "hello", "type": "STRING"}
        assert update_pipeline_resp.data["updatePipeline"]["parameters"][1] == {
            "name": "duration", "value": "0.1", "type": "FLOAT"}
        assert update_pipeline_resp.data["updatePipeline"]["tags"][0] == {
            "key": "author", "value": "opensean"}
        assert update_pipeline_resp.data["updatePipeline"]["tags"][1] == {
            "key": "package", "value": "kedro-graphql"}

    @pytest.mark.asyncio
    async def test_delete_pipeline(self,
                                   mock_app,
                                   mock_celery_session_app,
                                   celery_session_worker,
                                   mock_info_context,
                                   mock_text_in,
                                   mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "tags": [{"key": "author", "value": "opensean"}, {"key": "package", "value": "kedro-graphql"}]
                                                             }, "uniquePaths": None})

        pipeline_id = create_pipeline_resp.data["createPipeline"]["id"]

        delete_pipeline_resp = await mock_app.schema.execute(self.delete_pipeline_mutation,
                                                             variable_values={"id": pipeline_id})

        assert delete_pipeline_resp.errors is None

    @pytest.mark.asyncio
    async def test_pipeline_data_catalog_modified_with_log_datasets(self,
                                                                    mock_app,
                                                                    mock_celery_session_app,
                                                                    celery_session_worker,
                                                                    mock_info_context,
                                                                    mock_text_in,
                                                                    mock_text_out):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example00",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "text_out", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                                 ],
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",

                                                             }, "uniquePaths": None})

        query = """
    	  subscription {
          	pipeline(id:""" + '"' + str(create_pipeline_resp.data["createPipeline"]["id"]) + '"' + """) {
              id
              status
              result
            }
    	  }
        """
        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors
            if result.data["pipeline"]["status"] == "SUCCESS":
                break
        dataset_names = {ds.name for ds in mock_app.backend.read(
            create_pipeline_resp.data["createPipeline"]["id"]).data_catalog}

        assert "gql_meta" in dataset_names
        assert "gql_logs" in dataset_names

    @pytest.mark.asyncio
    async def test_pipeline_slicing(self,
                                    mock_app,
                                    mock_celery_session_app,
                                    celery_session_worker,
                                    mock_info_context,
                                    mock_text_in,
                                    mock_uppercased_txt,
                                    mock_reversed_txt,
                                    mock_timestamped_txt):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example01",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "uppercased", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_uppercased_txt)})},
                                                                                 {"name": "reversed", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_reversed_txt)})},
                                                                                 {"name": "timestamped", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_timestamped_txt)})}],
                                                                 "slices": {"slice": "NODE_NAMES", "args": ["uppercase_node", "reverse_node"]},
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                             }, "uniquePaths": None})

        query = """
    	  subscription {
          	pipeline(id:""" + '"' + str(create_pipeline_resp.data["createPipeline"]["id"]) + '"' + """) {
              id
              status
              result
            }
    	  }
        """
        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors
            if result.data["pipeline"]["status"] == "SUCCESS":
                break

        # Make sure only nodes specified in "slices" were run
        assert mock_app.backend.read(create_pipeline_resp.data["createPipeline"]
                                     ["id"]).status[-1].filtered_nodes == ["uppercase_node", "reverse_node"]
        create_pipeline_resp.errors is None

    @pytest.mark.asyncio
    async def test_pipeline_run_only_missing(self,
                                             mock_app,
                                             mock_celery_session_app,
                                             celery_session_worker,
                                             mock_info_context,
                                             mock_text_in,
                                             mock_uppercased_txt,
                                             mock_reversed_txt,
                                             mock_timestamped_txt):

        create_pipeline_resp = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                             variable_values={"pipeline": {
                                                                 "name": "example01",
                                                                 "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                                 {"name": "uppercased", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_uppercased_txt)})},
                                                                                 {"name": "reversed", "config": json.dumps(
                                                                                     {"type": "text.TextDataset", "filepath": str(mock_reversed_txt)})},
                                                                                 {"name": "timestamped", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_timestamped_txt)})}],
                                                                 "onlyMissing": True,
                                                                 "parameters": [{"name": "example", "value": "hello"},
                                                                                {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                                 "state": "READY",
                                                             }, "uniquePaths": None})
        query = """
    	  subscription {
          	pipeline(id:""" + '"' + str(create_pipeline_resp.data["createPipeline"]["id"]) + '"' + """) {
              id
              status
              result
            }
    	  }
        """
        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors
            if result.data["pipeline"]["status"] == "SUCCESS":
                break

        # Make sure only timestamp_node was run because the file does not exist (did not write to it in conftest.py)
        assert mock_app.backend.read(create_pipeline_resp.data["createPipeline"]
                                     ["id"]).status[-1].filtered_nodes == ["timestamp_node", "timestamp_partitions_node"]
        create_pipeline_resp.errors is None

    @pytest.mark.asyncio
    async def test_create_datasets(self,
                                   mock_app,
                                   mock_text_in,
                                   mock_text_out,
                                   mock_info_context):

        # create a staged pipeline
        response = await mock_app.schema.execute(self.create_pipeline_mutation,
                                                 variable_values={"pipeline": {
                                                     "name": "example01",
                                                     "dataCatalog": [{"name": "text_in", "config": json.dumps({"type": "text.TextDataset", "filepath": str(mock_text_in)})},
                                                                     {"name": "text_out", "config": json.dumps(
                                                                         {"type": "text.TextDataset", "filepath": str(mock_text_out)})}
                                                                     ],
                                                     "onlyMissing": True,
                                                     "parameters": [{"name": "example", "value": "hello"},
                                                                    {"name": "duration", "value": "0.1", "type": "FLOAT"}],
                                                     "state": "STAGED",
                                                 }, "uniquePaths": None})

        create_datasets_resp = await mock_app.schema.execute(self.create_datasets_mutation,
                                                             variable_values={"id": str(response.data["createPipeline"]["id"]),
                                                                              "datasets": [{"name": "text_in"}, {"name": "text_out"}],
                                                                              "expiresInSec": 3600})
        assert create_datasets_resp.errors is None
        assert create_datasets_resp.data["createDatasets"] is not None
        assert isinstance(create_datasets_resp.data["createDatasets"], list)
        assert isinstance(create_datasets_resp.data["createDatasets"][0], dict)
        assert len(create_datasets_resp.data["createDatasets"]) == 2

    @pytest.mark.asyncio
    async def test_create_datasets_not_staged(self,
                                              mock_app,
                                              mock_pipeline,
                                              mock_info_context):

        create_datasets_resp = await mock_app.schema.execute(self.create_datasets_mutation,
                                                             variable_values={"id": str(mock_pipeline.id),
                                                                              "datasets": [{"name": "text_in"}, {"name": "text_out"}],
                                                                              "expiresInSec": 3600})
        assert create_datasets_resp.errors is not None
        assert create_datasets_resp.data is None



import pytest
import asyncio
import time


class TestSchemaSubscriptions:
    @pytest.mark.asyncio
    async def test_pipeline(self, mock_app, mock_info_context, mock_pipeline):
        """Requires Redis to run.
        """

        query = """
    	  subscription {
          	pipeline(id:""" + '"' + str(mock_pipeline.id) + '"' + """) {
              id
              taskId
              status
              result
              timestamp
              traceback
            }
    	  }
        """
        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors

    @pytest.mark.asyncio
    async def test_pipeline_logs(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Requires Redis to run.
        This test runs two pipelines simultaneously to ensure logs messages are scoped
        to the correct pipeline.

        """

        query = """
    	  subscription {
          	pipelineLogs(id:""" + '"' + str(mock_pipeline.id) + '"' + """) {
              id
              message
              messageId
              taskId
              time
            }
    	  }
        """

        sub = await mock_app.schema.subscribe(query)

        async for result in sub:
            assert not result.errors
            assert result.data["pipelineLogs"]["id"] == str(mock_pipeline.id)
            assert result.data["pipelineLogs"]["taskId"] == str(
                mock_pipeline.status[-1].task_id)

        query2 = """
    	  subscription {
          	pipelineLogs(id:""" + '"' + str(mock_pipeline2.id) + '"' + """) {
              id
              message
              messageId
              taskId
              time
            }
    	  }
        """

        sub2 = await mock_app.schema.subscribe(query2)

        async for result in sub2:
            assert not result.errors
            assert result.data["pipelineLogs"]["id"] == str(mock_pipeline2.id)
            assert result.data["pipelineLogs"]["taskId"] == str(
                mock_pipeline2.status[-1].task_id)

    @pytest.mark.asyncio
    async def test_pipeline_subscription_non_blocking(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Test that pipeline subscriptions don't block the event loop.
        
        This test runs two pipeline subscriptions concurrently and verifies they execute
        in parallel rather than sequentially. If the subscriptions were blocking, they would
        run serially and take much longer.
        
        Requires Redis to run.
        """
        
        async def subscribe_to_pipeline(pipeline_id):
            """Helper to subscribe to a single pipeline and collect events."""
            query = """
              subscription {
                pipeline(id:""" + '"' + str(pipeline_id) + '"' + """) {
                  id
                  taskId
                  status
                  result
                  timestamp
                  traceback
                }
              }
            """
            events = []
            start_time = time.time()
            
            sub = await mock_app.schema.subscribe(query)
            async for result in sub:
                assert not result.errors
                events.append(result.data["pipeline"])
                
            elapsed = time.time() - start_time
            return {"pipeline_id": pipeline_id, "events": events, "elapsed": elapsed}
        
        # Run both subscriptions concurrently
        start_time = time.time()
        results = await asyncio.gather(
            subscribe_to_pipeline(mock_pipeline.id),
            subscribe_to_pipeline(mock_pipeline2.id)
        )
        total_elapsed = time.time() - start_time
        
        # Verify both subscriptions completed
        assert len(results) == 2
        assert all(len(r["events"]) > 0 for r in results)
        
        # Verify the subscriptions ran concurrently (not sequentially)
        # If they ran sequentially, total time would be sum of individual times
        # If they ran concurrently, total time should be close to the max of individual times
        individual_times_sum = sum(r["elapsed"] for r in results)
        max_individual_time = max(r["elapsed"] for r in results)
        
        # Total elapsed time should be much closer to max individual time than to sum
        # This indicates concurrent execution. Allow some overhead tolerance.
        assert total_elapsed < individual_times_sum * 0.75, \
            f"Subscriptions appear to be blocking. Total time ({total_elapsed:.2f}s) " \
            f"is too close to sequential execution time ({individual_times_sum:.2f}s). " \
            f"Expected closer to max individual time ({max_individual_time:.2f}s)."
        
        # Verify we got events for both pipelines
        pipeline_ids = {r["events"][0]["id"] for r in results}
        assert str(mock_pipeline.id) in pipeline_ids
        assert str(mock_pipeline2.id) in pipeline_ids

    @pytest.mark.asyncio
    async def test_pipeline_subscription_high_concurrency(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Test that many concurrent pipeline subscriptions don't block each other.
        
        This test simulates a more realistic scenario with multiple clients
        connecting simultaneously. If blocking occurs, we'll see significantly
        degraded performance.
        
        Requires Redis to run.
        """
        
        async def subscribe_and_count(pipeline_id, subscription_num):
            """Subscribe and count events received."""
            query = """
              subscription {
                pipeline(id:""" + '"' + str(pipeline_id) + '"' + """, interval: 0.1) {
                  id
                  status
                }
              }
            """
            count = 0
            start_time = time.time()
            
            sub = await mock_app.schema.subscribe(query)
            async for result in sub:
                assert not result.errors
                count += 1
                
            elapsed = time.time() - start_time
            return {"subscription": subscription_num, "count": count, "elapsed": elapsed}
        
        # Run 6 concurrent subscriptions (3 per pipeline)
        start_time = time.time()
        results = await asyncio.gather(
            subscribe_and_count(mock_pipeline.id, 1),
            subscribe_and_count(mock_pipeline.id, 2),
            subscribe_and_count(mock_pipeline.id, 3),
            subscribe_and_count(mock_pipeline2.id, 4),
            subscribe_and_count(mock_pipeline2.id, 5),
            subscribe_and_count(mock_pipeline2.id, 6),
        )
        total_elapsed = time.time() - start_time
        
        # All subscriptions should complete
        assert len(results) == 6
        assert all(r["count"] > 0 for r in results)
        
        # With true concurrency, total time should not scale linearly with number of subscriptions
        # If blocking, 6 subscriptions would take ~6x the time of 1 subscription
        # With concurrency, should take approximately the same as 1 subscription (plus overhead)
        individual_times_sum = sum(r["elapsed"] for r in results)
        max_individual_time = max(r["elapsed"] for r in results)
        
        # Total time should be MUCH less than sum (indicates parallel execution)
        # Allow for some overhead but should be closer to max than to sum
        assert total_elapsed < individual_times_sum * 0.5, \
            f"High concurrency test suggests blocking. Total time ({total_elapsed:.2f}s) " \
            f"is too high compared to sequential time ({individual_times_sum:.2f}s). " \
            f"With {len(results)} concurrent subscriptions, expected ~{max_individual_time:.2f}s."

    @pytest.mark.asyncio
    async def test_pipeline_logs_subscription_non_blocking(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Test that pipeline_logs subscriptions don't block the event loop.
        
        This test runs two pipeline_logs subscriptions concurrently and verifies they execute
        in parallel rather than sequentially. If the subscriptions were blocking, they would
        run serially and take much longer.
        
        Requires Redis to run.
        """
        
        async def subscribe_to_pipeline_logs(pipeline_id):
            """Helper to subscribe to a single pipeline's logs and collect messages."""
            query = """
              subscription {
                pipelineLogs(id:""" + '"' + str(pipeline_id) + '"' + """) {
                  id
                  message
                  messageId
                  taskId
                  time
                }
              }
            """
            messages = []
            start_time = time.time()
            
            sub = await mock_app.schema.subscribe(query)
            async for result in sub:
                assert not result.errors
                messages.append(result.data["pipelineLogs"])
                
            elapsed = time.time() - start_time
            return {"pipeline_id": pipeline_id, "messages": messages, "elapsed": elapsed}
        
        # Run both subscriptions concurrently
        start_time = time.time()
        results = await asyncio.gather(
            subscribe_to_pipeline_logs(mock_pipeline.id),
            subscribe_to_pipeline_logs(mock_pipeline2.id)
        )
        total_elapsed = time.time() - start_time
        
        # Verify both subscriptions completed
        assert len(results) == 2
        assert all(len(r["messages"]) > 0 for r in results)
        
        # Verify the subscriptions ran concurrently (not sequentially)
        # If they ran sequentially, total time would be sum of individual times
        # If they ran concurrently, total time should be close to the max of individual times
        individual_times_sum = sum(r["elapsed"] for r in results)
        max_individual_time = max(r["elapsed"] for r in results)
        
        # Total elapsed time should be much closer to max individual time than to sum
        # This indicates concurrent execution. Allow some overhead tolerance.
        assert total_elapsed < individual_times_sum * 0.75, \
            f"Pipeline logs subscriptions appear to be blocking. Total time ({total_elapsed:.2f}s) " \
            f"is too close to sequential execution time ({individual_times_sum:.2f}s). " \
            f"Expected closer to max individual time ({max_individual_time:.2f}s)."
        
        # Verify we got log messages for both pipelines with correct pipeline IDs
        pipeline_ids = {r["messages"][0]["id"] for r in results}
        assert str(mock_pipeline.id) in pipeline_ids
        assert str(mock_pipeline2.id) in pipeline_ids
        
        # Verify task IDs match the pipeline task IDs
        for result in results:
            pipeline_id = result["pipeline_id"]
            task_id = result["messages"][0]["taskId"]
            if str(pipeline_id) == str(mock_pipeline.id):
                assert task_id == str(mock_pipeline.status[-1].task_id)
            else:
                assert task_id == str(mock_pipeline2.status[-1].task_id)

    @pytest.mark.asyncio
    async def test_event_loop_responsiveness_during_subscriptions(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Test that the event loop remains responsive during active subscriptions.
        
        This test runs subscriptions and simultaneously checks if the event loop
        can handle other async tasks. If subscriptions block, other tasks will be delayed.
        
        Requires Redis to run.
        """
        
        # Track when simple tasks complete
        task_completion_times = []
        
        async def simple_task(task_num, delay=0.1):
            """A simple async task that should complete quickly if event loop is free."""
            start = time.time()
            await asyncio.sleep(delay)
            completion_time = time.time() - start
            task_completion_times.append({
                "task": task_num,
                "expected_delay": delay,
                "actual_delay": completion_time
            })
        
        async def subscribe_to_pipeline(pipeline_id):
            """Subscribe to pipeline events."""
            query = """
              subscription {
                pipeline(id:""" + '"' + str(pipeline_id) + '"' + """, interval: 0.1) {
                  id
                  status
                }
              }
            """
            count = 0
            sub = await mock_app.schema.subscribe(query)
            async for result in sub:
                assert not result.errors
                count += 1
            return count
        
        # Start subscriptions and simple tasks concurrently
        # If subscriptions block, simple tasks will take much longer than expected
        await asyncio.gather(
            subscribe_to_pipeline(mock_pipeline.id),
            subscribe_to_pipeline(mock_pipeline2.id),
            simple_task(1, 0.1),
            simple_task(2, 0.2),
            simple_task(3, 0.3),
            simple_task(4, 0.1),
            simple_task(5, 0.2),
        )
        
        # Verify all simple tasks completed
        assert len(task_completion_times) == 5
        
        # Check that simple tasks completed close to their expected time
        # If the event loop was blocked, delays would be much longer
        for task_info in task_completion_times:
            # Allow 50% overhead for task scheduling and other async operations
            # If blocking occurs, delays would be 2x-10x longer
            max_acceptable_delay = task_info["expected_delay"] * 1.5
            assert task_info["actual_delay"] < max_acceptable_delay, \
                f"Task {task_info['task']} took {task_info['actual_delay']:.3f}s " \
                f"when it should have taken ~{task_info['expected_delay']:.3f}s. " \
                f"This suggests the event loop was blocked by subscriptions."

    @pytest.mark.asyncio
    async def test_pipeline_logs_subscription_high_concurrency(self, mock_app, mock_info_context, mock_pipeline, mock_pipeline2):
        """Test high concurrency scenario with 6 concurrent log subscriptions.
        
        This is a more robust test that simulates realistic high-load conditions.
        If any blocking occurs, it will be more apparent with more concurrent connections.
        
        Requires Redis to run.
        """
        
        async def subscribe_to_pipeline_logs(pipeline_id, sub_num):
            """Helper to subscribe to a single pipeline's logs and collect messages."""
            query = """
              subscription {
                pipelineLogs(id:""" + '"' + str(pipeline_id) + '"' + """) {
                  id
                  message
                  messageId
                  taskId
                  time
                }
              }
            """
            messages = []
            start_time = time.time()
            
            sub = await mock_app.schema.subscribe(query)
            async for result in sub:
                assert not result.errors
                messages.append(result.data["pipelineLogs"])
                
            elapsed = time.time() - start_time
            return {
                "sub_num": sub_num,
                "pipeline_id": pipeline_id,
                "messages": messages,
                "elapsed": elapsed
            }
        
        # Run 6 subscriptions concurrently (3 for each pipeline)
        start_time = time.time()
        results = await asyncio.gather(
            subscribe_to_pipeline_logs(mock_pipeline.id, 1),
            subscribe_to_pipeline_logs(mock_pipeline.id, 2),
            subscribe_to_pipeline_logs(mock_pipeline.id, 3),
            subscribe_to_pipeline_logs(mock_pipeline2.id, 4),
            subscribe_to_pipeline_logs(mock_pipeline2.id, 5),
            subscribe_to_pipeline_logs(mock_pipeline2.id, 6),
        )
        total_elapsed = time.time() - start_time
        
        # Verify all subscriptions completed
        assert len(results) == 6
        assert all(len(r["messages"]) > 0 for r in results)
        
        # Calculate timing metrics
        individual_times_sum = sum(r["elapsed"] for r in results)
        max_individual_time = max(r["elapsed"] for r in results)
        avg_individual_time = individual_times_sum / len(results)
        
        # With 6 concurrent subscriptions, if they run truly in parallel,
        # total time should be close to max individual time.
        # If blocking occurs, total time approaches sum of all times.
        # Use stricter threshold for high concurrency test
        assert total_elapsed < individual_times_sum * 0.5, \
            f"High concurrency log subscription test shows blocking behavior. " \
            f"Total time: {total_elapsed:.2f}s, " \
            f"Sum of individual times: {individual_times_sum:.2f}s, " \
            f"Max individual time: {max_individual_time:.2f}s, " \
            f"Avg individual time: {avg_individual_time:.2f}s. " \
            f"Expected total time close to max ({max_individual_time:.2f}s), " \
            f"but got {total_elapsed:.2f}s which is too close to sequential execution."
        
        # Verify we got messages for both pipelines
        pipeline_ids = {r["messages"][0]["id"] for r in results}
        assert str(mock_pipeline.id) in pipeline_ids
        assert str(mock_pipeline2.id) in pipeline_ids
        
        # Verify task IDs are correct for each pipeline
        for result in results:
            pipeline_id = result["pipeline_id"]
            task_id = result["messages"][0]["taskId"]
            if str(pipeline_id) == str(mock_pipeline.id):
                assert task_id == str(mock_pipeline.status[-1].task_id)
            else:
                assert task_id == str(mock_pipeline2.status[-1].task_id)




from jinja2 import Environment, PackageLoader, select_autoescape
import json
from kedro.io import AbstractDataSet, DataCatalog, MemoryDataSet
from kedro.pipeline import Pipeline
from kedro.runner.runner import AbstractRunner
from pluggy import PluginManager
import re
import requests
import yaml




class ArgoWorkflowsRunner(AbstractRunner):
    """``ArgoWorkflowsRunner`` is an ``AbstractRunner`` implementation. It can be used
    to run pipelines on [argo workflows](https://argoproj.github.io/argo-workflows/).
    """
    ## NEED TO ADD TO kedro_graphq.config
    namespace = "default"
    host = "http://127.0.0.1:2746"
    endpoints = {"get_workflow": "/api/v1/workflows/{namespace}/{name}",
                 "create_workflow": "/api/v1/workflows/{namespace}",
                 "workflow_logs": "/api/v1/workflows/{namespace}/{name}/log"}
    image = "docker/whalesay:latest"
    package_name = "kedro-graphql"
    env = Environment(
        loader=PackageLoader("kedro_graphql.runner.argo"), autoescape=select_autoescape()
    )


    def create_default_data_set(self, ds_name: str) -> AbstractDataSet:
        """Factory method for creating the default data set for the runner.

        NOTE THIS SHOULD BE CHANGED TO SOMETHING S3 COMPATIBLE.

        Args:
            ds_name: Name of the missing data set
        Returns:
            An instance of an implementation of AbstractDataSet to be used
            for all unregistered data sets.

        """
        return MemoryDataSet()

    def _run(
        self,
        pipeline: Pipeline,
        catalog: DataCatalog,
        hook_manager: PluginManager = None,
        session_id: str = None,
    ) -> None:
        """The method implementing argo workflows pipeline running.
        Example logs output using this implementation:

         

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: The ``DataCatalog`` from which to fetch data.
            session_id: The id of the session.

        """
        nodes = pipeline.nodes
        self._logger.info(
            "Actual run will execute %s nodes:\n%s",
            len(nodes),
            pipeline.describe(),
        )
        self._logger.info("Checking inputs...")
        input_names = pipeline.inputs()
        missing_inputs = [
            input_name
            for input_name in input_names
            if not catalog._get_dataset(input_name).exists()
        ]
        if missing_inputs:
            raise KeyError(f"Datasets {missing_inputs} not found.")


        tasks = self.get_dependencies(pipeline.node_dependencies)

        template = self.env.get_template("argo_spec.tmpl")
        output = template.render(image=self.image, package_name=self.package_name, tasks=tasks)
        manifest =  yaml.safe_load(output)
        
        
        resp0 = self.create_workflow(manifest)

        ## stream argo logs to kedro logger
        self.workflow_logs(resp0["metadata"]["name"])

        resp1 = self.get_workflow(resp0["metadata"]["name"])

        self._logger.info("Completed " + resp1["status"]["progress"] + " tasks")

        if resp1["status"]["phase"] == "Failed":
            raise RuntimeError()
        
        ## need to check out status types and reconnect if still running, maybe while loop?
        ## also what happens when connection to argo server is lost during call
        ## to workflow_logs, how to restablish etc...

    def clean_name(self, name):
        return re.sub(r"[\W_]+", "-", name).strip("-")

    def create_workflow(self, manifest):
        """
        
        """
        endpoint = self.host + self.endpoints["create_workflow"].format(namespace = self.namespace)
        resp = requests.post(url = endpoint, json = {"workflow":manifest})
        return resp.json()

        
    def get_dependencies(self, dependencies):
        deps_dict = [
            {
                "node": node.name,
                "name": self.clean_name(node.name),
                "deps": [self.clean_name(val.name) for val in parent_nodes],
            }
            for node, parent_nodes in dependencies.items()
        ]
        return deps_dict

    def get_workflow(self, name):
        """
        
        """
        endpoint = self.host + self.endpoints["get_workflow"].format(namespace = self.namespace, name = name)
        resp = requests.get(url = endpoint)
        return resp.json()


    def workflow_logs(self, name):
        """
        
        Inspired by https://github.com/argoproj/argo-workflows/issues/4017


        """
        endpoint = self.host + self.endpoints["workflow_logs"].format(namespace = self.namespace, name = name)
        params = {
            'logOptions.container': 'main',
            'logOptions.follow': True
        }
        with requests.get(url=endpoint, params=params, stream=True) as resp:
            for _line in resp.iter_lines():
                self._logger.info(json.loads(_line.decode())["result"]["content"])






from importlib import import_module
from kedro_graphql.logs.logger import logger

UI_PLUGINS = {"FORMS": {},
              "DATA": {},
              "DASHBOARD": {},
              }


def discover_plugins(config):
    """Discover and import plugins based on the configuration.
    Args:
        config (dict): Configuration dictionary containing the imports.
    """
    for i in config["KEDRO_GRAPHQL_IMPORTS"]:
        import_module(i)


def ui_form(pipeline):
    """Register a UI form plugin for a specific pipeline.

    Args:
        pipeline (str): Name of the pipeline for which the form is registered.
    """

    def register_plugin(plugin_class):
        if UI_PLUGINS["FORMS"].get(pipeline, False):
            UI_PLUGINS["FORMS"][pipeline].append(plugin_class)
        else:
            UI_PLUGINS["FORMS"][pipeline] = [plugin_class]
        logger.info("registered ui_form plugin: " + str(plugin_class))
        return plugin_class

    return register_plugin


def ui_data(pipeline):
    """Register a UI data plugin for a specific pipeline.

    Args:
        pipeline (str): Name of the pipeline for which the data plugin is registered.
    """
    def register_plugin(plugin_class):
        if UI_PLUGINS["DATA"].get(pipeline, False):
            UI_PLUGINS["DATA"][pipeline].append(plugin_class)
        else:
            UI_PLUGINS["DATA"][pipeline] = [plugin_class]
        logger.info("registered ui_data plugin: " + str(plugin_class))
        return plugin_class

    return register_plugin


def ui_dashboard(pipeline):
    """Register a UI dashboard plugin for a specific pipeline.

    Args:
        pipeline (str): Name of the pipeline for which the dashboard plugin is registered.
    """
    def register_plugin(plugin_class):
        if UI_PLUGINS["DASHBOARD"].get(pipeline, False):
            UI_PLUGINS["DASHBOARD"][pipeline].append(plugin_class)
        else:
            UI_PLUGINS["DASHBOARD"][pipeline] = [plugin_class]
        logger.info("registered ui_dashboard plugin: " + str(plugin_class))
        return plugin_class

    return register_plugin



"""A python panel app for visualizing Kedro pipelines."""
import panel as pn
# from kedro_graphql.ui.components.template import KedroGraphqlMaterialTemplate, KedroGraphqlMaterialTemplateV2
from kedro_graphql.ui.components.template import KedroGraphqlMaterialTemplateV2
from importlib import import_module
import tempfile
import yaml

pn.extension(design='material', global_css=[
             ':root { --design-primary-color: black; }'])


# def template_factory(spec={}):
# """Factory function to create a Kedro GraphQL UI template.
##
# Args:
# spec (dict): The specification for the UI, containing configuration and pages.
# Returns:
# dict: A dictionary mapping the base URL to a function that builds the template.
# """
##
# def build_template():
# return KedroGraphqlMaterialTemplate(spec=spec)
##
# return {spec["panel_get_server_kwargs"]["base_url"]: build_template}


def build_app_factory(spec, page, config={}):
    """Factory function to create a Kedro GraphQL UI application.

    Args:
        spec (dict): The specification for the UI, containing configuration and pages.
        page (str): The name of the page to build.
        config (dict): Configuration dictionary for the page.
    Returns:
        pn.Column: A Panel column containing the specified page.
    """

    def build_app():
        if config.get("params", False):
            params = config["params"]
        else:
            params = {}
        return KedroGraphqlMaterialTemplateV2(spec=spec, main=config["module"](spec=spec, **params))

    return build_app


def start_ui(config={}, spec=""):
    """Start the Kedro GraphQL UI application.

    Args:
        config (dict): Configuration dictionary.
        spec (str): Path to the YAML specification file for the UI.
    """

    # load the UI yaml specification
    with open(spec) as stream:
        try:
            spec = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    # import components specified in UI spec for pages
    for key, value in spec["pages"].items():
        module_path, class_name = value["module"].rsplit(".", 1)
        module = __import__(module_path, fromlist=[class_name])
        module_class = getattr(module, class_name)
        spec["pages"][key]["module"] = module_class
        print("imported module: " +
              str(module_class) + " for page: " + str(key))

    # import additinal modules to enable plugin discovery
    # e.g. @gql_form, @gql_data, etc...
    imports = [i.strip() for i in spec["config"]["imports"]]
    for i in imports:
        import_module(i)

    import sh

    with tempfile.TemporaryDirectory() as tmpdirname:
        sh.kedro("viz", "build")
        sh.mv("build", tmpdirname + "/build")

        pn.config.reuse_sessions = True
        pn.config.admin = True
        pn.config.global_loading_spinner = True

        if spec["panel_get_server_kwargs"].get("static_dirs", None):
            spec["panel_get_server_kwargs"]["static_dirs"]["/pipeline/viz-build"] = str(
                tmpdirname + "/build")
        else:
            spec["panel_get_server_kwargs"]["static_dirs"] = {
                "/pipeline/viz-build": str(tmpdirname + "/build")}

        # pn.serve(template_factory(spec=spec), **spec["panel_get_server_kwargs"])
        apps = {}
        for page, config in spec["pages"].items():
            apps[page] = build_app_factory(spec, page, config)

        pn.serve(apps, **spec["panel_get_server_kwargs"])



import json

import pytest
from omegaconf import OmegaConf

from kedro_graphql.models import DataSet, DataSetInput, Parameter, ParameterInput, PipelineInput, TagInput
from .utilities import kedro_graphql_config
from pathlib import Path


class TestDataSet:

    config = kedro_graphql_config()

    def test_serialize(self):
        params = {"name": "text_in",
                  "config": json.dumps({"type": "text.TextDataset",
                                        "filepath": "/tmp/test_in.csv",
                                        "load_args": {"delimiter": "\t"},
                                        "save_args": {"delimiter": "\t"}})
                  }

        expected = {"text_in": {
            "type": "text.TextDataset",
            "filepath": "/tmp/test_in.csv",
            "load_args": {
                "delimiter": "\t"
            },
            "save_args": {
                "delimiter": "\t"
            }
        }
        }

        d = DataSet(**params)
        output = d.serialize()

        assert output == expected

    def test_does_exist_with_config(self, mock_text_in):
        params = {
            "name": "text_in",
            "config": f'{{"type": "text.TextDataset", "filepath": "{str(mock_text_in)}"}}'
        }

        d = DataSet(**params)
        assert d.exists() == True

    def test_does_not_exist_with_config(self):
        params = {
            "name": "text_in",
            "config": '{"type": "text.TextDataset", "filepath": "/tmp/does_not_exist.csv"}'
        }

        d = DataSet(**params)
        assert d.exists() == False

    def test_partitions(self):
        d = DataSet(name="test_partitioned_dataset", config=json.dumps(
            {"type": "partitions.PartitionedDataset",
             "path": str(Path("src/tests/data/partitioned_dataset/").resolve()),
             "filename_suffix": ".txt",
             "dataset": {"type": "text.TextDataset"}}
        ))

        assert d.partitions() == ["part-0001", "part-0002"]


class TestParameterInput:

    def test_create_from_dict(self):
        correct = {
            "a": "b",
            "c": 0,
            "d": True,
            "e": 0.1, }

        incorrect = {**correct, "f": [1, 2]}

        params_input_list = ParameterInput.create(correct)

        assert len(params_input_list) == 4
        assert any(p.name == "a" and p.value == "b" for p in params_input_list)

        with pytest.raises(ValueError):
            ParameterInput.create(incorrect)


class TestParameter:

    def test_serialize_string(self):
        params = {
            "name": "delimiter",
                    "value": "\t",
                    "type": "string"
        }

        expected = {
            "delimiter": "\t"
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_int(self):
        params = {
            "name": "delimiter",
                    "value": "1",
                    "type": "integer"
        }

        expected = {
            "delimiter": 1
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_int_exception(self):
        params = {
            "name": "delimiter",
                    "value": "0.1",
                    "type": "integer"
        }

        p = Parameter(**params)
        try:
            output = p.serialize()
        except ValueError as e:
            assert True

    def test_serialize_float(self):
        params = {
            "name": "delimiter",
                    "value": "0.1",
                    "type": "float"
        }

        expected = {
            "delimiter": 0.1
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_float_exception(self):
        params = {
            "name": "delimiter",
                    "value": "hello",
                    "type": "float"
        }

        p = Parameter(**params)
        try:
            output = p.serialize()
        except ValueError as e:
            assert True

    def test_serialize_bool(self):

        params = {
            "name": "delimiter",
                    "value": "true",
                    "type": "boolean"
        }

        expected = {
            "delimiter": True
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

        params = {
            "name": "delimiter",
                    "value": "True",
                    "type": "boolean"
        }

        expected = {
            "delimiter": True
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

        params = {
            "name": "delimiter",
                    "value": "false",
                    "type": "boolean"
        }

        expected = {
            "delimiter": False
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

        params = {
            "name": "delimiter",
                    "value": "False",
                    "type": "boolean"
        }

        expected = {
            "delimiter": False
        }

        p = Parameter(**params)
        output = p.serialize()
        assert output == expected

    def test_serialize_bool_exception(self):

        params = {
            "name": "delimiter",
                    "value": "rue",
                    "type": "boolean"
        }

        p = Parameter(**params)
        try:
            output = p.serialize()
        except ValueError as e:
            assert True

    def test_dotlist_notation_to_omega_conf(self):
        """
        Tests serialized Parameter objects with dotlist notation names can be converted to OmegaConf
        which is used to construct the DataCatalog with the add_feed_dict method in tasks.py.
        """
        parameter_inputs = [{"name": "example", "value": "hello", "type": "string"},
                            {"name": "duration", "value": "0.1", "type": "float"},
                            {"name": "model_options.model_params.learning_date",
                                "value": "2023-11-01", "type": "string"},
                            {"name": "model_options.model_params.training_date",
                                "value": "2023-11-01", "type": "string"},
                            {"name": "model_options.model_params.data_ratio",
                                "value": "14", "type": "float"},
                            {"name": "data_options.step_size",
                                "value": "123123", "type": "float"},
                            ]

        parameters = [Parameter.decode(p) for p in parameter_inputs]

        serialized_parameters = {}

        for p in parameters:
            serialized_parameters.update(p.serialize())

        parameters_dotlist = [f"{key}={value}" for key,
                              value in serialized_parameters.items()]
        conf_parameters = OmegaConf.from_dotlist(parameters_dotlist)
        kedro_parameters = {"parameters": conf_parameters}

        assert kedro_parameters == {
            'parameters': {
                'example': 'hello',
                'duration': 0.1,
                'model_options': {
                    'model_params': {
                        'learning_date': '2023-11-01',
                        'training_date': '2023-11-01',
                        'data_ratio': 14
                    }
                },
                'data_options': {
                    'step_size': 123123
                }
            }
        }

        params_dotlist = [f"params:{key}={value}" for key,
                          value in serialized_parameters.items()]
        kedro_params = OmegaConf.from_dotlist(params_dotlist)

        assert kedro_params == {
            'params:example': 'hello',
            'params:duration': 0.1,
            'params:model_options': {
                'model_params': {
                    'learning_date': '2023-11-01',
                    'training_date': '2023-11-01',
                    'data_ratio': 14
                }
            },
            'params:data_options': {
                'step_size': 123123
            }
        }

    def test_pipeline_encode_as_input(self, mock_pipeline_staged):
        """
        Tests the Pipeline.encode(encoder="input") method returns a PipelineInput object
        """
        result = mock_pipeline_staged.encode(encoder="input")
        assert isinstance(result, PipelineInput)
        assert result.name == mock_pipeline_staged.name
        assert len(result.data_catalog) == len(mock_pipeline_staged.data_catalog)
        assert len(result.parameters) == len(mock_pipeline_staged.parameters)
        assert len(result.tags) == len(mock_pipeline_staged.tags)


class TestDataSetInput:

    def test_encode(self):
        """
        Tests the DataSet.encode() method returns a DataSetInput object
        """
        dataset = DataSetInput(
            name="text_in",
            config=json.dumps({
                "type": "text.TextDataset",
                "filepath": "/tmp/test_in.csv",
                "load_args": {"delimiter": "\t"},
                "save_args": {"delimiter": "\t"}
            })
        )

        result = dataset.encode(encoder="graphql")
        assert isinstance(result, dict)
        assert result["name"] == dataset.name
        assert result["config"] == dataset.config



import logging
from importlib import import_module

logger = logging.getLogger("kedro")


def init_runner(runner_import_path: str, **runner_kwargs):
    module, class_name = runner_import_path.rsplit(".", 1)

    module = import_module(module)
    runner_cls = getattr(module, class_name)
    logger.info("using runner " + str(runner_cls))
    return runner_cls(**runner_kwargs)



import ast
import json
import uuid

from bson.objectid import ObjectId
from pymongo import MongoClient

from kedro_graphql.models import Pipeline

from .base import BaseBackend


class MongoBackend(BaseBackend):

    def __init__(self, uri=None, db=None, collection="pipelines"):
        self.client = MongoClient(uri)
        self.db = self.client[db]
        self.collection = collection

    def startup(self, **kwargs):
        """Startup hook."""
        print("Connected to the MongoDB database!")

    def shutdown(self, **kwargs):
        """Shutdown hook."""
        self.client.close()

    def list(self, cursor: uuid.UUID = None, limit=10, filter="", sort=""):
        query = {}
        if len(filter) > 0:
            filter = json.loads(filter)
            query = filter
        if cursor is not None:
            query.update({'_id': {'$gte': ObjectId(cursor)}})

        if sort:
            try:
                sort = ast.literal_eval(sort)
                # Validate that sort is a list of tuples like [('created_at', -1)]
                if isinstance(sort, list) and all(isinstance(i, tuple) and len(i) == 2 for i in sort):
                    raw = self.db[self.collection].find(query).sort(sort).limit(limit)
                else:
                    raise ValueError(
                        "Sort parameter should be a list of tuples like [('field', order)]")
            except (ValueError, SyntaxError) as e:
                raise ValueError(f"Invalid sort parameter format: {e}")
        else:
            raw = self.db[self.collection].find(query).limit(limit)

        results = []
        for r in raw:
            r["id"] = str(r["_id"])
            p = Pipeline.decode(r)
            results.append(p)
        return results

    def read(self, id: uuid.UUID = None, task_id: str = None):
        """Load a pipeline by id or task_id"""
        if task_id:
            r = self.db[self.collection].find_one(
                {"status": {"$elemMatch": {"task_id": task_id}}})
        else:
            r = self.db[self.collection].find_one({"_id": ObjectId(id)})

        if r:
            r["id"] = str(r["_id"])
            p = Pipeline.decode(r)
            return p
        else:
            return None

    def create(self, pipeline: Pipeline):
        """Save a pipeline"""
        values = pipeline.encode()
        values.pop("id")  # we dont have an id yet, we will get it after insert
        created = self.db[self.collection].insert_one(values)
        created = self.db[self.collection].find_one({"_id": created.inserted_id})
        created["id"] = str(created["_id"])
        p = Pipeline.decode(created)
        return p

    def update(self, pipeline: Pipeline = None):
        """Update a pipeline"""
        id = ObjectId(pipeline.id)
        filter = {'_id': id}
        values = pipeline.encode()
        values.pop("id")  # we dont want to update the id
        newvalues = {"$set": values}
        self.db[self.collection].update_one(filter, newvalues)

        p = self.read(id=id)

        return p

    def delete(self, id: uuid.UUID = None):
        """Delete a pipeline using id"""
        self.db[self.collection].delete_one({"_id": ObjectId(id)})
        return id



"""
This is a boilerplate pipeline 'event00'
generated using Kedro 0.19.11
"""
from kedro_graphql.client import KedroGraphqlClient
from kedro_graphql.models import PipelineInput
from cloudevents.pydantic.v1 import CloudEvent
from cloudevents.conversion import from_json
import json
import asyncio
import tempfile
from omegaconf.dictconfig import DictConfig
from omegaconf import OmegaConf


def create_pipeline_input(id: str, event: CloudEvent) -> PipelineInput:
    """
    Create a PipelineInput object based on the provided ID and CloudEvent.

    Args:
        id (str): The ID of the pipeline to which this node belongs.
        event (CloudEvent): A CloudEvent object containing event data.

    Returns:
        PipelineInput: A PipelineInput object ready for use in the pipeline.
    """

    tmp = tempfile.gettempdir()
    mock_text_in = tmp + "/text_in.txt"
    mock_text_out = tmp + "/text_out.txt"

    with open(mock_text_in, "w") as f:
        f.write("Hello, world!")

    with open(mock_text_out, "w") as f:
        f.write("Goodbye, world!")

    input_dict = {"type": "text.TextDataset", "filepath": str(mock_text_in)}
    output_dict = {"type": "text.TextDataset", "filepath": str(mock_text_out)}
    return PipelineInput(**{
        "parent": id,
        "name": "example00",
        "state": "STAGED",
        "data_catalog": [{"name": "text_in", "config": json.dumps(input_dict)},
                         {"name": "text_out", "config": json.dumps(output_dict)}],
        "parameters": [{"name": "example", "value": "hello"},
                       {"name": "duration", "value": "0", "type": "FLOAT"}],
        "tags": [{"key": "author", "value": "opensean"},
                 {"key": "package", "value": "kedro-graphql"}]
    })


# i dont know why kedro serializes the event as a DictConfig, it should be string
# need to look into the details of the MemoryDataset and parameters handling
def run_example00_via_api(id: str, event: DictConfig) -> dict:
    """
    Example node that parse a cloudevent passed as json
    and triggers the example00 pipeline via the GraphQL API.

    Args:
        id (str): The ID of the pipeline to which this node belongs.
        event (str): A JSON string representing a CloudEvent.

    Returns:
        dict: A dictionary representation of the created pipeline.
    """
    event = from_json(CloudEvent, json.dumps(OmegaConf.to_container(event)))

    client = KedroGraphqlClient(uri_graphql="http://localhost:5000/graphql",
                                uri_ws="ws://localhost:5000/graphql")

    pipeline_input = create_pipeline_input(id, event)

    pipeline = asyncio.run(client.create_pipeline(pipeline_input))

    asyncio.run(client.close_sessions())
    return pipeline.encode(encoder="dict")



import pytest
from kedro_graphql.schema import PipelineSanitizer, DataSetConfigException
from kedro_graphql.models import PipelineInput
import json


class TestSchemaExtensions:

    @pytest.mark.asyncio
    async def test_pipeline_sanitizer_mask_filepaths(self, mock_pipeline_staged):
        """
         Test that the kedro_graphql.schema.PipelineSanitizer correctly masks filepaths in a PipelineInput's data catalog.
        """
        result = mock_pipeline_staged.encode(encoder="input")
        assert isinstance(result, PipelineInput)

        masks = [{"prefix": "./data/", "mask": "./REDACTED/"}]
        masked = PipelineSanitizer.mask_filepaths(result, masks)
        for d in masked.data_catalog:
            c = json.loads(d.config)
            if c.get("filepath"):
                assert not c["filepath"].startswith("./data/")
                assert c["filepath"].startswith("./REDACTED/")

    @pytest.mark.asyncio
    async def test_pipeline_sanitizer_unmask_filepaths(self, mock_pipeline_staged):
        """
         Test that the kedro_graphql.schema.PipelineSanitizer correctly unmasks filepaths in a PipelineInput's data catalog.
        """
        result = mock_pipeline_staged.encode(encoder="input")
        assert isinstance(result, PipelineInput)

        masks = [{"prefix": "./data/", "mask": "./REDACTED/"}]
        masked = PipelineSanitizer.mask_filepaths(result, masks)
        unmasked = PipelineSanitizer.unmask_filepaths(masked, masks)
        for d in unmasked.data_catalog:
            c = json.loads(d.config)
            if c.get("filepath"):
                assert not c["filepath"].startswith("./REDACTED/")
                assert c["filepath"].startswith("./data/")

    @pytest.mark.asyncio
    async def test_pipeline_sanitizer_sanitize_filepaths(self, mock_pipeline_staged):
        """
         Test that the kedro_graphql.schema.PipelineSanitizer correctly validates filepaths in a PipelineInput's data catalog against allowed prefixes.
        """
        result = mock_pipeline_staged.encode(encoder="input")
        assert isinstance(result, PipelineInput)
        allowed_roots = ["./data/"]
        sanitized = PipelineSanitizer.sanitize_filepaths(result, allowed_roots)

        allowed_roots = ["./tmp/"]
        try:
            sanitized = PipelineSanitizer.sanitize_filepaths(result, allowed_roots)
        except DataSetConfigException as e:
            assert "filepath ./data/01_raw/text_in.csv not allowed" in str(e)



import json
from typing import List
import jwt
from kedro.io import AbstractDataset
from strawberry.types import Info
from ..models import DataSet, SignedUrl, SignedUrls, SignedUrlField
from .base import SignedUrlProvider
from datetime import datetime, timedelta
from urllib.parse import urlencode
from pathlib import Path
from ..permissions import get_permissions
from ..logs.logger import logger
from ..config import load_config
from ..exceptions import DataSetError

CONFIG = load_config()
logger.debug("configuration loaded by {s}".format(s=__name__))

PERMISSIONS_CLASS = get_permissions(CONFIG.get("KEDRO_GRAPHQL_PERMISSIONS"))
logger.info("{s} using permissions class: {d}".format(s=__name__, d=PERMISSIONS_CLASS))


class LocalFileProvider(SignedUrlProvider):
    """
    Implementation of SignedUrlProvider for a local file system.
    """

    @staticmethod
    def sign_url(filepath: str | Path, expires_in_sec: int, url: str) -> dict:
        """
        Generate a signed URL for a local file.

        Args:
            filepath (pathlib.Path | str): The local file path.
            expires_in_sec (int): The number of seconds the signed URL should be valid for.
            url (str): The base URL for the signed URL.

        Returns:
            dict: A signed URL for the local file e.g. {"url": "http://example.com/download?token=abc123", "fields": {"token": "abc123"}}
        """
        if isinstance(filepath, str):
            path = Path(filepath).resolve()
        else:
            path = filepath.resolve()

        payload = {
            "filepath": str(path),
            "exp": int((datetime.now() + timedelta(seconds=expires_in_sec)).timestamp()),
            "iat": int(datetime.now().timestamp())
        }

        token = jwt.encode(payload, CONFIG["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY"],
                           algorithm=CONFIG["KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM"])

        return SignedUrl(url=url, file=path.name, fields=[SignedUrlField(name="token", value=token)])

    @staticmethod
    def read(info: Info, dataset: DataSet, expires_in_sec: int, partitions: list | None = None) -> SignedUrl | SignedUrls:
        """
        Get a signed URL for reading a dataset.

        Args:
            info (Info): Strawberry GraphQL Info object.
            dataset (DataSet): The dataset to read.
            expires_in_sec (int): The expiration time in seconds.
            partitions (list | None): Optional list of partitions in a PartitionedDataset.

        Returns:
            str | List[str]: A signed URL for reading the dataset.

        Raises:
            DataSetError: If signed URLs for the dataset type are not supported.
        """
        c = dataset.parse_config()

        if c["type"] in ["partitions.PartitionedDataset"]:

            protocol, path = dataset.parse_path()
            path = Path(path)
            if not partitions or len(partitions) == 0:
                # if no partitions provided,
                # need the list of files in the partitioned dataset to generate signed URLs for each partition
                files = AbstractDataset.from_config(dataset.name, c).load()
                files = [str(path / f) + c.get("filename_suffix", "") for f in files]
            else:
                files = [str(path / partition) + c.get("filename_suffix", "")
                         for partition in partitions]
            signed_urls = []
            for file in files:
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, filepath={file}, protocol=file, expires_in_sec={expires_in_sec}")

                signed = LocalFileProvider.sign_url(
                    file, expires_in_sec, f"{CONFIG['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL']}/download")

                query = urlencode({"token": signed.get_field_value("token")})
                signed.url = f"{CONFIG['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL']}/download?{query}"
                signed_urls.append(signed)
            return SignedUrls(urls=signed_urls)

        elif c["type"] in ["partitions.IncrementalDataset"]:
            raise DataSetError(
                "Signed URLs for reading IncrementalDatasets are not supported.")
        else:
            protocol, filepath = dataset.parse_filepath()
            filepath = Path(filepath)

            logger.info(
                f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=read_dataset, dataset={dataset.name}, filepath={filepath}, protocol=file, expires_in_sec={expires_in_sec}")
            signed = LocalFileProvider.sign_url(
                filepath, expires_in_sec, f"{CONFIG['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL']}/download")

            query = urlencode({"token": signed.get_field_value("token")})
            signed.url = f"{CONFIG['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL']}/download?{query}"
            return signed

    @staticmethod
    def create(info: Info, dataset: DataSet, expires_in_sec: int, partitions: list | None = None) -> SignedUrl | SignedUrls:
        """
        Get a signed URL for creating a dataset.

        Args:
            info (Info): Strawberry GraphQL Info object.
            dataset (DataSet): The dataset to create a signed URL for.
            expires_in_sec (int): The expiration time in seconds.
            partitions (list | None): Optional list of partitions in a PartitionedDataset.

        Returns:
            dict | List[dict]: A signed URL for creating the dataset.

        Raises:
            ValueError: If partitions are not provided for a PartitionedDataset.
            DataSetError: If signed URLs for the dataset type are not supported.
        """
        c = dataset.parse_config()
        if c["type"] in ["partitions.PartitionedDataset"]:
            if not partitions or len(partitions) == 0:
                raise ValueError(
                    "Partitions must be provided for PartitionedDataset")

            signed_urls = []
            protocol, path = dataset.parse_path()
            path = Path(path)
            for partition in partitions:
                filepath = str(path / partition) + c.get("filename_suffix", "")
                logger.info(
                    f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, dataset={dataset.name}, filepath={str(filepath)}, protocol=file, expires_in_sec={expires_in_sec}")

                signed_urls.append(LocalFileProvider.sign_url(
                    filepath, expires_in_sec, f"{CONFIG['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL']}/upload"))

            return SignedUrls(urls=signed_urls)
        elif c["type"] in ["partitions.IncrementalDataset"]:
            raise DataSetError(
                "Signed URLs for creating IncrementalDatasets are not supported.")
        else:
            protocol, filepath = dataset.parse_filepath()
            filepath = Path(filepath)
            logger.info(
                f"user={PERMISSIONS_CLASS.get_user_info(info)['email']}, action=create_dataset, dataset={dataset.name}, filepath={str(filepath)}, protocol=file, expires_in_sec={expires_in_sec}")

            return LocalFileProvider.sign_url(filepath, expires_in_sec, f"{CONFIG['KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL']}/upload")



import os
import tempfile

from dotenv import dotenv_values
import json
import yaml
from importlib import import_module
from .logs.logger import logger
import copy


defaults = {
    "KEDRO_GRAPHQL_APP": "kedro_graphql.asgi.KedroGraphQL",
    "KEDRO_GRAPHQL_APP_DESCRIPTION": "A tool for serving kedro projects as a GraphQL API",
    "KEDRO_GRAPHQL_APP_TITLE": "Kedro GraphQL API",
    "KEDRO_GRAPHQL_BACKEND": "kedro_graphql.backends.mongodb.MongoBackend",
    "KEDRO_GRAPHQL_BROKER": "redis://localhost",
    "KEDRO_GRAPHQL_CELERY_RESULT_BACKEND": "redis://localhost",
    "KEDRO_GRAPHQL_CLIENT_URI_GRAPHQL": "http://localhost:5000/graphql",
    "KEDRO_GRAPHQL_CLIENT_URI_WS": "ws://localhost:5000/graphql",
    "KEDRO_GRAPHQL_CONF_SOURCE": None,
    "KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS": [],
    "KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS": [],
    "KEDRO_GRAPHQL_DEPRECATIONS_DOCS": None,
    "KEDRO_GRAPHQL_ENV": "local",
    "KEDRO_GRAPHQL_EVENTS_CONFIG": None,
    "KEDRO_GRAPHQL_IMPORTS": ["kedro_graphql.plugins.plugins"],
    "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS": ["./data", "/var", "/tmp"],
    "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_ALGORITHM": "HS256",
    "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_JWT_SECRET_KEY": "my-secret-key",
    "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_SERVER_URL": "http://localhost:5000",
    "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS": ["./data"],
    "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_MAX_FILE_SIZE_MB": 10,
    "KEDRO_GRAPHQL_LOG_PATH_PREFIX": None,
    "KEDRO_GRAPHQL_LOG_TMP_DIR": tempfile.TemporaryDirectory().name,
    "KEDRO_GRAPHQL_MONGO_DB_COLLECTION": "pipelines",
    "KEDRO_GRAPHQL_MONGO_DB_NAME": "pipelines",
    "KEDRO_GRAPHQL_MONGO_URI": "mongodb://root:example@localhost:27017/",
    "KEDRO_GRAPHQL_PERMISSIONS": "kedro_graphql.permissions.IsAuthenticatedAlways",
    "KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP": {
        # "KEDRO_GRAPHQL_PERMISSIONS": "kedro_graphql.permissions.IsAuthenticatedXForwardedEmail",
        "EXTERNAL_GROUP_NAME": "admin"
    },
    "KEDRO_GRAPHQL_PERMISSIONS_ROLE_TO_ACTION_MAP": {
        "admin": ["create_pipeline",
                  "read_pipeline",
                  "read_pipelines",
                  "update_pipeline",
                  "delete_pipeline",
                  "read_pipeline_template",
                  "read_pipeline_templates",
                  "create_dataset",
                  "read_dataset",
                  "subscribe_to_events",
                  "subscribe_to_logs",
                  "create_event"]
    },
    "KEDRO_GRAPHQL_PROJECT_VERSION": "None",
    "KEDRO_GRAPHQL_ROOT_PATH": "",
    "KEDRO_GRAPHQL_RUNNER": "kedro.runner.SequentialRunner",
    # "KEDRO_GRAPHQL_RUNNER": "kedro_graphql.runner.argo.ArgoWorkflowsRunner",
    "KEDRO_GRAPHQL_SIGNED_URL_MAX_EXPIRES_IN_SEC": 43200,
    "KEDRO_GRAPHQL_SIGNED_URL_PROVIDER": "kedro_graphql.signed_url.s3_provider.S3Provider",
}


def load_api_spec():
    """Load API configuration from yaml file."""
    # load the UI yaml specification
    spec = os.environ.get("KEDRO_GRAPHQL_API_SPEC", None)
    if spec:
        with open(spec) as stream:
            try:
                conf = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                logger.error(str(exc))
                return {}

        # import additinal modules to enable plugin discovery
        # e.g. @gql_form, @gql_data, etc...
        imports = conf["config"]["imports"]
        if isinstance(imports, str):
            # Handle comma-separated string
            imports = [i.strip() for i in imports.split(',') if i.strip()]
        elif isinstance(imports, list):
            # Already a list, just ensure strings are stripped
            imports = [i.strip() for i in imports if i.strip()]

        for i in imports:
            import_module(i)

        config = {"_".join(["KEDRO_GRAPHQL", k.upper()]): v for k,
                  v in conf["config"].items()}

        return config
    else:
        logger.debug("No API spec file found. Using default configuration.")
        return {}


cli_config = {}  # placeholder for CLI config


def env_var_parser(config):
    # Parse JSON strings for complex data types
    json_fields = [
        "KEDRO_GRAPHQL_EVENTS_CONFIG",
        "KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP",
        "KEDRO_GRAPHQL_PERMISSIONS_ROLE_TO_ACTION_MAP",
        "KEDRO_GRAPHQL_DATASET_FILEPATH_ALLOWED_ROOTS",
        "KEDRO_GRAPHQL_DATASET_FILEPATH_MASKS",
    ]

    # Fields that can be either JSON arrays, comma-separated strings, or lists
    list_fields = [
        "KEDRO_GRAPHQL_IMPORTS",
        "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_DOWNLOAD_ALLOWED_ROOTS",
        "KEDRO_GRAPHQL_LOCAL_FILE_PROVIDER_UPLOAD_ALLOWED_ROOTS",
    ]

    for field in json_fields:
        if config.get(field, None) and isinstance(config[field], str):
            try:
                config[field] = json.loads(config[field])
            except json.JSONDecodeError:
                logger.warning(f"Failed to parse JSON for {field}: {config[field]}")

    # Handle list fields. Can be string (comma-separated) or list or JSON array
    for field in list_fields:
        if field in config:
            value = config[field]
            if isinstance(value, str):
                if value.strip():
                    # Try to parse as JSON first
                    try:
                        config[field] = json.loads(value)
                    except json.JSONDecodeError:
                        # If not JSON, treat as comma-separated string
                        config[field] = [i.strip()
                                         for i in value.split(',') if i.strip()]
                else:
                    # Empty string should become empty list
                    config[field] = []
    return config


def load_config(cli_config=cli_config):
    """Load configuration from the environment variables and API spec.

    Configuration precedence (highest to lowest):
    1. YAML API spec
    2. CLI flags
    3. Environment variables
    4. .env file
    5. Defaults
    """

    config = {
        **copy.deepcopy(defaults),  # defaults (lowest precedence)
        **dotenv_values(".env"),  # .env file
        **os.environ,  # environment variables
        **cli_config,  # CLI flags (higher precedence than env vars)
        **load_api_spec(),  # YAML API spec (highest precedence)
    }
    config = env_var_parser(config)  # special parsing for any environment variables

    return config



import json
import os
from datetime import datetime
from pathlib import Path
from unittest.mock import patch
import tempfile

import pytest
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project

from kedro_graphql.asgi import KedroGraphQL
from kedro_graphql.models import (
    DataSet,
    Parameter,
    Pipeline,
    PipelineStatus,
    State,
    Tag,
)
from kedro_graphql.tasks import run_pipeline
from fastapi.middleware.cors import CORSMiddleware
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from kedro_graphql.client import KedroGraphqlClient
from multiprocessing import Process
import uvicorn
from pathlib import Path
from kedro_graphql.asgi import KedroGraphQL
import multiprocessing as mp
import tempfile
import pytest_asyncio
from .utilities import kedro_graphql_config


if mp.get_start_method(allow_none=True) != "spawn":
    mp.set_start_method("spawn")


@pytest.fixture(scope="session")
def kedro_session():
    bootstrap_project(Path.cwd())
    return KedroSession.create()


def start_server(port=5000, config={}):

    with tempfile.TemporaryDirectory() as tmp:
        with tempfile.TemporaryDirectory() as tmp2:
            bootstrap_project(Path.cwd())
            session = KedroSession.create()
            app = KedroGraphQL(kedro_session=session, config=config)
            app.config["KEDRO_GRAPHQL_LOG_PATH_PREFIX"] = tmp
            app.config["KEDRO_GRAPHQL_LOG_TMP_DIR"] = tmp2
            # print("Starting Kedro GraphQL server with config: ", app.config)
            app.add_middleware(
                CORSMiddleware,
                allow_origins=["*"],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )
            uvicorn.run(app,
                        host="localhost",
                        port=port)


@pytest.fixture(scope="session")
def mock_server():
    config = kedro_graphql_config()
    proc = Process(target=start_server, args=(), kwargs={
                   "port": 5000,
                   "config": config})
    proc.start()
    yield
    proc.terminate()


@pytest.fixture(scope="session")
def mock_server_5001():
    config = kedro_graphql_config()
    config["KEDRO_GRAPHQL_PERMISSIONS"] = "kedro_graphql.permissions.IsAuthenticatedXForwardedRBAC"
    config["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"] = {
        "test_group": "admin"
    }
    # set env variables so other modules will load config correctly
    os.environ["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"] = json.dumps(
        config["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"])
    os.environ["KEDRO_GRAPHQL_PERMISSIONS"] = config["KEDRO_GRAPHQL_PERMISSIONS"]

    proc = Process(target=start_server, args=(), kwargs={
                   "port": 5001,
                   "config": config})
    proc.start()
    yield
    proc.terminate()


@pytest.fixture(scope="session")
def mock_server_5002():
    config = kedro_graphql_config()
    config["KEDRO_GRAPHQL_PERMISSIONS"] = "kedro_graphql.permissions.IsAuthenticatedXForwardedEmail"
    config["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"] = {
        "test_group": "admin"
    }
    # set env variables so other modules will load config correctly
    os.environ["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"] = json.dumps(
        config["KEDRO_GRAPHQL_PERMISSIONS_GROUP_TO_ROLE_MAP"])
    os.environ["KEDRO_GRAPHQL_PERMISSIONS"] = config["KEDRO_GRAPHQL_PERMISSIONS"]

    proc = Process(target=start_server, args=(), kwargs={
                   "port": 5002,
                   "config": config})
    proc.start()
    yield
    proc.terminate()


@pytest_asyncio.fixture
async def mock_client(mock_server):

    client = KedroGraphqlClient(uri_graphql="http://localhost:5000/graphql",
                                uri_ws="ws://localhost:5000/graphql")
    yield client
    await client.close_sessions()


@pytest.fixture(scope="session")
def mock_app(kedro_session):
    config = kedro_graphql_config()
    with tempfile.TemporaryDirectory() as tmp:
        with tempfile.TemporaryDirectory() as tmp2:
            app = KedroGraphQL(kedro_session=kedro_session, config=config)
            app.config["KEDRO_GRAPHQL_LOG_PATH_PREFIX"] = tmp
            app.config["KEDRO_GRAPHQL_LOG_TMP_DIR"] = tmp2

            yield app


@pytest.fixture(scope='session')
def celery_config():
    return {
        'broker_url': 'redis://',
        'result_backend': 'redis://',
        'result_extened': True,
        'worker_send_task_events': True,
        'task_send_sent_event': True,
        'task_store_eager_result': True,
        'task_always_eager': False,
        'task_ignore_result': False,
        'imports': ["kedro_graphql.tasks"]

    }


@pytest.fixture(scope='session')
def mock_celery_session_app(mock_app, mock_info_context, celery_session_app):
    celery_session_app.kedro_graphql_backend = mock_app.backend
    celery_session_app.kedro_graphql_schema = mock_app.schema
    celery_session_app.kedro_graphql_config = mock_app.config
    return celery_session_app


@pytest.fixture(scope="session")
def celery_worker_parameters():
    return {"without_heartbeat": False}


@pytest.fixture(scope="session")
def mock_info_context(mock_app):

    class Request():
        app = mock_app
        headers = {}

    with patch("strawberry.types.Info.context", {"request": Request()}) as m:
        yield m


# refer to https://docs.pytest.org/en/7.1.x/how-to/tmp_path.html for info on tmp_path fixture
@pytest.fixture
def mock_text_in(tmp_path):
    text_in = tmp_path / "text_in.txt"
    text_in.write_text("hello")
    return text_in


@pytest.fixture
def mock_text_out(tmp_path):
    text_out = tmp_path / "text_out.txt"
    text_out.write_text("good bye")
    return text_out


@pytest.fixture
def mock_text_in_tsv(tmp_path):
    text = tmp_path / "text_in.tsv"
    text.write_text(
        "Some parameter\tOther parameter\tLast parameter\nCONST\t123456\t12.45")
    return text


@pytest.fixture
def mock_text_out_tsv(tmp_path):
    text = tmp_path / "text_out.tsv"
    text.write_text(
        "Some parameter\tOther parameter\tLast parameter\nCONST\t123456\t12.45")
    return text


@pytest.fixture
def mock_uppercased_txt(tmp_path):
    text = tmp_path / "uppercased.txt"
    text.write_text("HELLO")
    return text


@pytest.fixture
def mock_reversed_txt(tmp_path):
    text = tmp_path / "reversed.txt"
    text.write_text("OLLEH")
    return text


@pytest.fixture
def mock_timestamped_txt(tmp_path):
    text = tmp_path / "timestamped.txt"
    # Don't write to this file. We need an missing file to test the run "only_missing" mutation
    return text


@pytest.fixture
def mock_pipeline(mock_celery_session_app,
                  celery_session_worker,
                  depends_on_current_app,
                  mock_app,
                  tmp_path,
                  mock_text_in,
                  mock_text_out):

    inputs = [{"name": "text_in", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": str(mock_text_in)})}]
    outputs = [{"name": "text_out", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": str(mock_text_out)})}]
    parameters = [{"name": "example", "value": "hello"}]
    tags = [{"key": "author", "value": "opensean"}, {
        "key": "package", "value": "kedro-graphql"}]

    p = Pipeline(
        name="example00",
        data_catalog=[DataSet(**i) for i in inputs] + [DataSet(**o) for o in outputs],
        parameters=[Parameter(**p) for p in parameters],
        tags=[Tag(**p) for p in tags],
        status=[PipelineStatus(state=State.READY,
                               runner=mock_app.config["KEDRO_GRAPHQL_RUNNER"],
                               session=mock_app.kedro_session.session_id,
                               started_at=datetime.now(),
                               task_name=str(run_pipeline))]
    )

    p.created_at = datetime.now()
    p = mock_app.backend.create(p)

    serial = p.serialize()

    result = run_pipeline.apply_async(kwargs={"id": str(p.id),
                                              "name": "example00",
                                              "data_catalog": serial["data_catalog"],
                                              "parameters": serial["parameters"],
                                              "runner": mock_app.config["KEDRO_GRAPHQL_RUNNER"]},
                                      countdown=0.1)

    print(f'Starting {p.name} pipeline with task_id: ' + str(result.id))
    p.status[-1].task_id = result.id
    p = mock_app.backend.update(p)
    return p


@pytest.fixture
def mock_pipeline_staged(mock_app):

    inputs = [{"name": "text_in", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": "./data/01_raw/text_in.csv"})}]
    outputs = [{"name": "text_out", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": "./data/01_raw/text_out.csv"})}]
    parameters = [{"name": "example", "value": "hello"}]
    tags = [{"key": "author", "value": "opensean"}, {
        "key": "package", "value": "kedro-graphql"}]

    p = Pipeline(
        name="example00",
        data_catalog=[DataSet(**i) for i in inputs] + [DataSet(**o) for o in outputs],
        parameters=[Parameter(**p) for p in parameters],
        tags=[Tag(**p) for p in tags],
        status=[PipelineStatus(state=State.STAGED,
                               runner=mock_app.config["KEDRO_GRAPHQL_RUNNER"],
                               session=mock_app.kedro_session.session_id,
                               started_at=datetime.now(),
                               task_name=str(run_pipeline))]
    )

    p.created_at = datetime.now()
    p = mock_app.backend.create(p)
    return p


@pytest.fixture
def mock_pipeline2(mock_app, tmp_path, mock_text_in, mock_text_out):

    inputs = [{"name": "text_in", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": str(mock_text_in)})}]
    outputs = [{"name": "text_out", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": str(mock_text_out)})}]
    parameters = [{"name": "example", "value": "hello"}]
    tags = [{"key": "author", "value": "opensean"}, {
        "key": "package", "value": "kedro-graphql"}]

    p = Pipeline(
        name="example00",
        data_catalog=[DataSet(**i) for i in inputs] + [DataSet(**o) for o in outputs],
        parameters=[Parameter(**p) for p in parameters],
        tags=[Tag(**p) for p in tags],
        status=[PipelineStatus(state=State.READY,
                               runner=mock_app.config["KEDRO_GRAPHQL_RUNNER"],
                               session=mock_app.kedro_session.session_id,
                               started_at=datetime.now(),
                               task_name=str(run_pipeline))]
    )

    p.created_at = datetime.now()
    p = mock_app.backend.create(p)

    serial = p.serialize()

    result = run_pipeline.apply_async(kwargs={"id": str(p.id),
                                              "name": "example00",
                                              "data_catalog": serial["data_catalog"],
                                              "parameters": serial["parameters"],
                                              "runner": mock_app.config["KEDRO_GRAPHQL_RUNNER"]},
                                      countdown=0.1)

    print(f'Starting {p.name} pipeline with task_id: ' + str(result.id))
    p.status[-1].task_id = result.id
    p = mock_app.backend.update(p)
    return p


@pytest.fixture
def mock_pipeline_no_task(mock_app, mock_text_in, mock_text_out):

    inputs = [{"name": "text_in", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": str(mock_text_in)})}]
    outputs = [{"name": "text_out", "config": json.dumps(
        {"type": "text.TextDataset", "filepath": str(mock_text_out)})}]
    parameters = [{"name": "example", "value": "hello"}]
    tags = [{"key": "author", "value": "opensean"}, {
        "key": "package", "value": "kedro-graphql"}]

    p = Pipeline(
        name="example00",
        data_catalog=[DataSet(**i) for i in inputs] + [DataSet(**o) for o in outputs],
        parameters=[Parameter(**p) for p in parameters],
        tags=[Tag(**p) for p in tags]
    )

    p.status.append(PipelineStatus(state=State.READY,
                                   runner=mock_app.config["KEDRO_GRAPHQL_RUNNER"],
                                   session=mock_app.kedro_session.session_id,
                                   started_at=datetime.now(),
                                   task_name=str(run_pipeline)))

    p.created_at = datetime.now()
    return p


@pytest.fixture(autouse=True)
def delete_pipeline_collection(mock_app):
    # Will be executed before the first test
    yield
    # Will be executed after the last test
    mock_app.backend.db[mock_app.config["KEDRO_GRAPHQL_MONGO_DB_NAME"]].drop()



import pytest

from cloudevents.http import CloudEvent
from cloudevents.conversion import to_structured
from requests import Request, Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry


class TestASGI:

    def event_data(self):
        attributes = {"id": "1234",
                      "type": "com.example.event",
                      "source": "example.com"}
        data = {"key": "value"}
        return CloudEvent(attributes, data)

    def event_post(self):
        retry = Retry(
            total=5,
            backoff_factor=0.5,
            status_forcelist=[429, 500, 502, 503, 504],
        )

        adapter = HTTPAdapter(max_retries=retry)
        s = Session()
        s.mount('http://', adapter)

        headers, body = to_structured(self.event_data())

        url = "http://localhost:5000/event/"
        req = Request('POST', url, data=body, headers=headers)
        req_prepped = req.prepare()
        # print("REQUEST:", req_prepped.headers, req_prepped.body)
        return s.send(req_prepped)

    @pytest.mark.asyncio
    async def test_post_event(self,
                              mock_server,
                              mock_celery_session_app,
                              celery_session_worker,):

        resp = self.event_post()
        assert resp.status_code == 200
        resp = resp.json()
        # print("RESPONSE JSON:", resp)
        assert len(resp) == 1
        pipe = resp[0]
        assert "name" in pipe
        assert pipe["name"] == "event00"
        assert pipe["status"][-1]['state'] in ["STARTED", "READY"]



import panel as pn
import param
from kedro_graphql.ui.decorators import UI_PLUGINS
from kedro_graphql.ui.components.pipeline_detail import PipelineDetail
from kedro_graphql.ui.components.pipeline_monitor import PipelineMonitor
from kedro_graphql.ui.components.pipeline_viz import PipelineViz
from kedro_graphql.ui.components.pipeline_retry import PipelineRetry
from kedro_graphql.ui.components.pipeline_cloning import PipelineCloning
from kedro_graphql.ui.components.data_catalog_explorer import DataCatalogExplorer


class PipelineDashboardFactory(pn.viewable.Viewer):
    """
    A factory for building dashboards for Kedro pipelines using registered @ui_dashboard plugins.
    This component allows users to select a dashboard for a specific pipeline and build the dashboard dynamically.

    Attributes:
        id (str): The ID of the pipeline to build the dashboard for.
        pipeline (str): The name of the pipeline for which the dashboard is built.
        options (list): A list of available dashboards for the selected pipeline.
        spec (dict): The specification for the UI, including configuration and pages.
        dashboard_name (str): The name of the selected dashboard.
        dataset_map (dict): A mapping of Kedro dataset types to their corresponding rendering Panel components.
    """

    id = param.String(default="")
    pipeline = param.String(default="")
    options = param.List(default=[])
    spec = param.Dict(default={})
    dashboard_name = param.String(default=None)
    dataset_map = param.Dict(default={})

    def __init__(self, **params):
        super().__init__(**params)

        pn.state.location.sync(
            self, {"id": "id", "pipeline": "pipeline", "dashboard_name": "dashboard_name"})

        if not self.id or not self.pipeline or not self.spec:
            self._content = pn.Column(
                pn.pane.Markdown(
                    "**Missing required parameters: `id`, `pipeline`, or `spec`.**"),
                sizing_mode="stretch_width"
            )
        else:
            # Show loading spinner until the dashboard is built
            self._content = pn.Column(
                pn.indicators.LoadingSpinner(value=True, width=50, height=50),
                pn.pane.Markdown("Fetching Data..."),
                sizing_mode='stretch_width'
            )

            # Ensures build after panel server is fully loaded to avoid race conditions
            pn.state.onload(self.build_dashboard)

    def build_default_dashboard(self, p):
        """
        Builds the default dashboard for a Kedro pipeline, including monitoring, detail, and visualization components
        registered to the pipeline using the @ui_data plugin.

        Args:
            p (Pipeline): The Kedro pipeline for which the dashboard is built.

        Returns:
            pn.Tabs: A panel containing tabs for monitoring, detail, and visualization of the pipeline.
        """

        monitor = PipelineMonitor(spec=self.spec, pipeline=p)
        detail = PipelineDetail(pipeline=p)
        viz = PipelineViz(pipeline=p.name, spec=self.spec)
        retry = PipelineRetry(client=self.spec["config"]["client"], pipeline=p)
        cloning = PipelineCloning(client=self.spec["config"]["client"], pipeline=p)
        explorer = DataCatalogExplorer(
            spec=self.spec,
            pipeline=p,
            dataset_map=self.dataset_map
        )
        tabs = pn.Tabs(dynamic=False)
        tabs.append(("Catalog", explorer))
        tabs.append(("Monitor", monitor))
        tabs.append(("Detail", detail))
        tabs.append(("Viz", viz))
        tabs.append(("Retry", retry))
        tabs.append(("Cloning", cloning))
        if UI_PLUGINS["DATA"].get(self.pipeline, None):
            for d in UI_PLUGINS["DATA"][self.pipeline]:
                data = d(id=self.id, spec=self.spec, pipeline=p)
                tabs.append((data.title, data))
        return tabs

    def build_custom_dashboard(self, p):
        """
        Builds a custom dashboard for a Kedro pipeline using a registered @ui_dashboard plugin.

        Args:
            p (Pipeline): The Kedro pipeline for which the dashboard is built.

        Returns:
            panel.viewable.Viewer: An instance of the custom dashboard class.
        """

        dash = None
        for d in UI_PLUGINS["DASHBOARD"][self.pipeline]:
            if d.__name__ == self.dashboard_name:
                dash = d
        dash = dash(id=self.id, pipeline=p, spec=self.spec)
        return dash

    async def build_dashboard(self):
        """
        Builds the dashboard for the selected pipeline and dashboard name.
        This method checks if a custom dashboard is registered for the pipeline and builds it accordingly.
        If no custom dashboard is registered, it builds the default dashboard with monitoring, detail, and visualization components.
        """

        p = await self.spec["config"]["client"].read_pipeline(id=self.id)

        if UI_PLUGINS["DASHBOARD"].get(self.pipeline, None):
            for f in UI_PLUGINS["DASHBOARD"][self.pipeline]:
                self.options.append(f.__name__)
            select = pn.widgets.Select.from_param(
                self.param.dashboard_name,
                name='Select a dashboard',
                options=self.options,
                value=self.param.dashboard_name
            )
            self.dashboard_name = self.options[0]
            custom_dashboard = self.build_custom_dashboard(p)
            self._content[:] = pn.Column(
                pn.Row(select),
                pn.Row(custom_dashboard)
            )
        else:
            self.options = []
            default_dashboard = self.build_default_dashboard(p)
            self._content[:] = pn.Row(default_dashboard)

    def __panel__(self):
        return self._content
